use std::fs::File;

use serde::Deserialize;

#[derive(Deserialize, Debug)]
struct ColorEntry {
    color: (u8, u8, u8),
    symbol: String,
}

#[derive(Deserialize, Debug)]
struct Mapping(Vec<ColorEntry>);

impl Mapping {
    fn load(path: impl AsRef<std::path::Path>) -> Result<Mapping, Box<dyn std::error::Error>> {
        Ok(serde_yaml::from_reader(File::open(path)?)?)
    }

    fn lookup(&self, color: (u8, u8, u8)) -> &str {
        let found = self.0.iter().find(|entry| entry.color == color);
        match found {
            Some(entry) => &entry.symbol,
            None => panic!("Unable to find entry for {:?}", color),
        }
    }
}


struct Image {
    width: u32,
    height: u32,
    data: Vec<(u8, u8, u8)>,
}

impl Image {
    fn load(path: impl AsRef<std::path::Path>) -> Result<Image, Box<dyn std::error::Error>> {
        let decoder = png::Decoder::new(File::open(path)?);
        let mut reader = decoder.read_info()?;
        let mut buf = vec![0u8; reader.output_buffer_size()];
        let info = reader.next_frame(buf.as_mut_slice())?;

        let mut data = Vec::new();

        let skip = match info.color_type {
            png::ColorType::Rgb => 3,
            png::ColorType::Rgba => 4,
            other => panic!("Unhandled color type: {:?}", other),
        };

        let rs = buf.iter().step_by(skip);
        let gs = buf.iter().skip(1).step_by(skip);
        let bs = buf.iter().skip(2).step_by(skip);

        for ((r, g), b) in rs.zip(gs).zip(bs) {
            data.push((*r, *g, *b))
        }

        Ok(Image {
            width: info.width,
            height: info.height,
            data
        })
    }
}

#[derive(Deserialize)]
struct Config {
    grid_every: u32,
    line_weight: f64,
    major_line_weight: f64,
    grid_size: f64,
}

impl Config {
    fn scale(&self, n: u32) -> f64 {
        n as f64 * self.grid_size
    }
}

impl Default for Config {
    fn default() -> Config {
        Config {
            grid_every: 10,
            line_weight: 1.0,
            major_line_weight: 3.0,
            grid_size: 24.0,
        }
    }
}

fn main() -> Result<(), Box<dyn std::error::Error>> {
    let img = Image::load("sample.png")?;
    let mapping = Mapping::load("colors.yaml")?;
    let mut x = 0;
    let mut y = 0;

    let config = Config::default();

    let surf = cairo::SvgSurface::new(
        config.scale(img.width),
        config.scale(img.height),
        Some("output.svg"),
    )?;

    let ctx = cairo::Context::new(&surf)?;
    ctx.set_source_rgb(1.0, 1.0, 1.0);
    ctx.paint()?;

    let layout = pangocairo::functions::create_layout(&ctx).unwrap();
    let font = pango::FontDescription::from_string("Fira Sans 12");
    layout.set_width(2880);
    layout.set_font_description(Some(&font));

    ctx.set_source_rgb(0.0, 0.0, 0.0);
    for i in 0..=img.width {
        ctx.move_to(config.scale(i), 0.0);
        ctx.line_to(config.scale(i), config.scale(img.height));
        ctx.set_line_width(if i % config.grid_every == 0 { config.major_line_weight } else { config.line_weight });
        ctx.stroke()?;
    }
    for j in 0..=img.height {
        ctx.move_to(0.0, config.scale(j));
        ctx.line_to(config.scale(img.width), config.scale(j));
        ctx.set_line_width(if j % config.grid_every == 0 { config.major_line_weight } else { config.line_weight });
        ctx.stroke()?;
    }

    ctx.set_source_rgb(0.0, 0.0, 0.0);

    for (r, g, b) in img.data.iter() {
        let l = mapping.lookup((*r, *g, *b));

        x += 1;
        if x == img.width {
            x = 0;
            y += 1;
        }

        if l != "" {
            ctx.move_to(config.scale(x - 1) + 5.0, config.scale(y) + 3.0);
            layout.set_text(l);
            pangocairo::functions::show_layout(&ctx, &layout);
        }
    }

    surf.finish();

    Ok(())
}
